package com.inteligr8.git;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.Semaphore;

import org.eclipse.jgit.api.CloneCommand;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.ResetCommand.ResetType;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.api.errors.InvalidRemoteException;
import org.eclipse.jgit.api.errors.RefNotFoundException;
import org.eclipse.jgit.api.errors.TransportException;
import org.eclipse.jgit.lib.Constants;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.util.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class LocalRepositoryCache {
	
	private static final LocalRepositoryCache INSTANCE = new LocalRepositoryCache();
	
	public static LocalRepositoryCache getInstance() {
		return INSTANCE;
	}
	
	
	
	private final Logger logger = LoggerFactory.getLogger(LocalRepositoryCache.class);
	
	private final LRUExpiringHashMap<String, CachedGit> cachedGits = new LRUExpiringHashMap<>(30);
	private final Map<String, Semaphore> gitUrlSemaphores = new HashMap<>();
	final File cacheDirectory = new File(System.getProperty("java.io.tmpdir"), "git");
	private int simultaneousThreadsPerGitRepo = 1;
	
	private LocalRepositoryCache() {
		this.cacheDirectory.mkdir();
		this.cachedGits.addListener(new RepositoryCacheMapListener<CachedGit>());
	}
	
	/**
	 * This is helpful for the tweaking performance based on the environment of
	 * execution.  A value of 1 disables concurrency which results in the
	 * persistence of Git Repositories between executions and disables cache
	 * expiration.  This is great for development, testing, and command line
	 * usage.  It is not recommended with web service frontends.
	 */
	public void setSimultaneousThreadsPerGitRepository(int threads) {
		this.simultaneousThreadsPerGitRepo = threads;
	}
	
	@Override
	protected void finalize() throws Throwable {
		try {
			this.destroy();
		} finally { 
			super.finalize();
		}
	}
	
	private void destroy() {
    	if (this.logger.isDebugEnabled())
    		this.logger.debug("clearing local repo cache");
		this.clear();
	}
	
	public CachedGit acquire(String url) throws InterruptedException, InvalidRemoteException, RefNotFoundException, TransportException, GitAPIException {
		return this.acquire(url, null, null);
	}
	
	public CachedGit acquire(String url, CredentialsProvider creds) throws InterruptedException, InvalidRemoteException, RefNotFoundException, TransportException, GitAPIException {
		return this.acquire(url, creds, null);
	}
	
	public CachedGit acquire(String url, String branch) throws InterruptedException, InvalidRemoteException, RefNotFoundException, TransportException, GitAPIException {
		return this.acquire(url, null, branch);
	}
	
	/**
	 * 
	 * @param url A URL for the Git Repository.
	 * @param creds A username/password for the Git Repository.
	 * @param branch A branch name of the branch in the Git Repository to start working on.
	 * @return A JGit object.
	 * @throws InterruptedException The JVM was interrupted while the thread was waiting for another request on the same Git Repository
	 * @throws InvalidRemoteException The URL refers to nothing or an invalid Git Repository
	 * @throws RefNotFoundException The branch does not exist in the Git Repository represented by the URL
	 * @throws TransportException A disk or network error or hiccup occurred.
	 * @throws GitAPIException A Git API, library, or protocol related issue occurred
	 */
	public CachedGit acquire(String url, CredentialsProvider creds, String branch)
	throws InterruptedException, InvalidRemoteException, RefNotFoundException, TransportException, GitAPIException {
    	if (this.logger.isTraceEnabled())
    		this.logger.trace("acquire('" + url + "', " + creds + ", '" + branch + "')");
    	
    	Semaphore semaphore = null;
    	synchronized (this) {
	    	semaphore = this.gitUrlSemaphores.get(url);
	    	if (semaphore == null)
	    		this.gitUrlSemaphores.put(url, semaphore = new Semaphore(this.simultaneousThreadsPerGitRepo));
    	}
    	
    	File gitRepoDirectory = null;
    	
    	semaphore.acquire();
    	try {
			CachedGit git = this.cachedGits.remove(url);
			if (git == null) {
				String directoryBaseName;
				if (this.simultaneousThreadsPerGitRepo == 1)
					directoryBaseName = new GitUrl(url).getFullyQualifiedRepositoryName().replace('/', '_');
				else directoryBaseName = UUID.randomUUID().toString();
				
				gitRepoDirectory = new File(this.cacheDirectory, directoryBaseName + ".git");
		    	if (this.logger.isDebugEnabled())
		    		this.logger.debug("Git directory cache for clone: " + gitRepoDirectory);
				
		    	if (gitRepoDirectory.exists()) {
			    	if (this.logger.isInfoEnabled())
			    		this.logger.info("Git Repository already exists; reusing: " + gitRepoDirectory);
		    		git = creds != null ? new CredentialedGit(Git.open(gitRepoDirectory), creds) : new ExtendedGit(Git.open(gitRepoDirectory));
		    	}
			}
			
			if (git == null) {
				CloneCommand clone = new CloneCommand()
		    			.setURI(url)
		    			.setDirectory(gitRepoDirectory);
				if (branch != null)
					clone.setBranch(branch);
				
		    	if (this.logger.isDebugEnabled())
		    		this.logger.debug("Cloning Git repository: " + url);
		    	git = creds != null ? new CredentialedGit(clone, creds) : new ExtendedGit(clone);
		    	if (this.logger.isInfoEnabled())
		    		this.logger.info("Cloned Git Repository: " + new GitUrl(git.getRemoteUrl("origin")).getFullyQualifiedRepositoryName());
			} else {
		    	if (this.logger.isDebugEnabled())
		    		this.logger.debug("resetting Git");
				git.reset().setMode(ResetType.HARD).call();
				git.fetch().call();
				
				if (branch != null && !branch.equals(git.getRepository().getBranch())) {
			    	if (this.logger.isDebugEnabled())
			    		this.logger.debug("switching Git branches: " + branch);
			    	if (git.getRepository().exactRef(Constants.R_HEADS + branch) == null) {
			    		git.checkout().setName(branch).setStartPoint("origin/" + branch).setCreateBranch(true).call();
			    	} else {
			    		git.checkout().setName(branch).call();
			    	}
				}
	
		    	if (this.logger.isDebugEnabled())
		    		this.logger.debug("updating Git branch: " + git.getRepository().getBranch());
				git.pull().call();
			}

	    	return git;
    	} catch (URISyntaxException use) {
    		semaphore.release();
    		throw new IllegalArgumentException("The Git Repository URL is not properly formatted", use);
    	} catch (IOException ie) {
    		semaphore.release();
    		throw new TransportException("A I/O issue occurred", ie);
    	} catch (GitAPIException gae) {
    		semaphore.release();
    		throw gae;
    	}
	}
	
	public void release(CachedGit git) {
		if (this.logger.isTraceEnabled())
    		this.logger.trace("release('" + git.getRepository().getIdentifier() + "')");
		
		try {
			String url = git.getRemoteUrl("origin").toString();
			
			synchronized (this) {
				Semaphore semaphore = this.gitUrlSemaphores.get(url);
				if (semaphore != null)
					semaphore.release();
			}
			
			this.cachedGits.put(url, git);
		} catch (GitAPIException gae) {
			if (this.logger.isDebugEnabled())
				this.logger.debug("A Git repository was never released", gae);
			this.logger.warn("A Git repository was never released, potentially blocking all future processing by this tool until a restart: " + git.toString());
		}
	}
	
	public void clear() {
		if (this.logger.isTraceEnabled())
    		this.logger.trace("clear()");
		
		this.cachedGits.clear();
		
		for (Semaphore semaphore : this.gitUrlSemaphores.values())
			; // FIXME interrupt all the threads; releasing semaphores will not be enough
		this.gitUrlSemaphores.clear();
	}
	
	private void expunge(Git git) {
		File gitDir = git.getRepository().getDirectory();
		File workingTreeDir = git.getRepository().getWorkTree();
		git.getRepository().close();
		
		try {
			if (this.logger.isDebugEnabled())
				this.logger.debug("deleting: " + gitDir);
			FileUtils.delete(gitDir, FileUtils.RECURSIVE);
		} catch (IOException ie) {
			this.logger.warn("Failed to delete a git directory: " + gitDir);
			if (this.logger.isDebugEnabled())
				this.logger.debug(ie.getMessage(), ie);
		}

		try {
			if (this.logger.isDebugEnabled())
				this.logger.debug("deleting: " + workingTreeDir);
			FileUtils.delete(workingTreeDir, FileUtils.RECURSIVE);
		} catch (IOException ie) {
			this.logger.warn("Failed to delete a git directory: " + workingTreeDir);
			if (this.logger.isDebugEnabled())
				this.logger.debug(ie.getMessage(), ie);
		}
		
		if (this.logger.isInfoEnabled())
			this.logger.info("Deleted Git Repository: " + gitDir);
	}
	
	
	
	private class RepositoryCacheMapListener<T extends CachedGit> implements ExpiringMapListener<String, T> {

		private final Logger logger = LoggerFactory.getLogger(LocalRepositoryCache.class);
		
		@Override
		public void accessed(Entry<String, T> entry) {
		}
		
		@Override
		public void added(Entry<String, T> entry) {
			// a clean one or one returned after being previously removed
		}
		
		@Override
		public void expired(Entry<String, T> entry	) {
			if (this.logger.isTraceEnabled())
	    		this.logger.trace("expired('" + entry.getKey() + "', '" + entry.getValue().getRepository().getIdentifier() + "')");
			expunge(entry.getValue());
		}
		
		@Override
		public void removed(Entry<String, T> entry) {
			// expected to be removed only temporarily...for use elsewhere; do not close
		}
		
		@Override
		public void cleared(Entry<String, T> entry) {
			if (this.logger.isTraceEnabled())
	    		this.logger.trace("cleared('" + entry.getKey() + "', '" + entry.getValue().getRepository().getIdentifier() + "')");
			expunge(entry.getValue());
		}
		
	}

}
